#include <types.h>
#include <alloc.h>
#include <spinlock.h>
#include <arch/amd64/vm.h>

#define BOOL(x) (!(!(x)))
#define BitMsk(bit) (1 << (bit))
#define BitNset(arg,bit) ((arg) |= BitMsk(bit))
#define BitNclr(arg,bit) ((arg) &= ~BitMsk(bit))
#define BitNtst(arg,bit) BOOL((arg) & BitMsk(bit))


#define KHEAP_CHUNK_COUNT 5
#define KHEAP_SIZE (32*1024*1024)

#define KHEAP_CHUNK_FREE 0
#define KHEAP_CHUNK_1 (sizeof(size_t)*8 / 1) //64 / 1 = 64 (64b)
#define KHEAP_CHUNK_2 (sizeof(size_t)*8 / 2) // 64 / 2 = 32 (128b)
#define KHEAP_CHUNK_4 (sizeof(size_t)*8 / 4) // 64 / 4 = 16 (256b)
#define KHEAP_CHUNK_8 (sizeof(size_t)*8 / 8) // 64 / 8 = 8 (512b)
#define KHEAP_CHUNK_16 (sizeof(size_t)*8 / 16) // 64 / 16 = 4 (1024b)

#define KHEAP_CHUNK_1_SIZE (PAGESIZE / KHEAP_CHUNK_1)
#define KHEAP_CHUNK_2_SIZE (PAGESIZE / KHEAP_CHUNK_2)
#define KHEAP_CHUNK_4_SIZE (PAGESIZE / KHEAP_CHUNK_4)
#define KHEAP_CHUNK_8_SIZE (PAGESIZE / KHEAP_CHUNK_8)
#define KHEAP_CHUNK_16_SIZE (PAGESIZE / KHEAP_CHUNK_16)

#define KHEAP_CHUNK_1_INDEX 0
#define KHEAP_CHUNK_2_INDEX 1
#define KHEAP_CHUNK_4_INDEX 2
#define KHEAP_CHUNK_8_INDEX 3
#define KHEAP_CHUNK_16_INDEX 4



typedef struct kheap_mcb_t
{

	uint8  type;
	size_t bitmap;

	struct kheap_mcb_t *next;
} kheap_mcb_t;

struct kheap_mcb_t *kheap_start;
struct kheap_mcb_t *kheap_first_mcb;
struct kheap_mcb_t *kheap_free_stack;
struct kheap_mcb_t *kheap_stacks[KHEAP_CHUNK_COUNT];

int kheap_chunk_sizes[KHEAP_CHUNK_COUNT];
uint32 kheap_chunk_allocs[KHEAP_CHUNK_COUNT];
uint32 kheap_chunk_new_mcb_allocs[KHEAP_CHUNK_COUNT];

struct spinlock_t kheap_lock;

void kheap_show_stats()
{
int i;
	
	printk("kheap_chunk_allocs:\n");
	for ( i = 0 ; i < KHEAP_CHUNK_COUNT;i++)
		printk("#%i - %i allocs\n",i,kheap_chunk_allocs[i]);

	printk("kheap_chunk_new_mcb_allocs:\n");
	for ( i = 0 ; i < KHEAP_CHUNK_COUNT;i++)
		printk("#%i - %i allocs\n",i,kheap_chunk_new_mcb_allocs[i]);
}


void kheap_push_free_stack(struct kheap_mcb_t *pp)
{

	if(pp->next)
		panic("kheap_push_free_stack() page is already on a stack page=%p\n",pp);

	pp->next = (struct kheap_mcb_t*)kheap_free_stack;
	kheap_free_stack = (struct kheap_mcb_t*) pp;
}

struct kheap_mcb_t *kheap_pop_free_stack()
{
struct kheap_mcb_t *pp;

	if(kheap_free_stack)
	{
	
		pp = (struct kheap_mcb_t*)kheap_free_stack;
		kheap_free_stack = (struct kheap_mcb_t*)pp->next;

	}
	else
		return NULL;
	
	pp->next = NULL;
	return (kheap_mcb_t*)pp;
}


struct kheap_mcb_t *kheap_pop_stack(struct kheap_mcb_t **stack)
{
struct kheap_mcb_t *p;

	if(*stack)
	{
		p = (struct kheap_mcb_t*)*stack;
		*stack = (struct kheap_mcb_t *)p->next;
	}
	else
		return NULL;

	p->next = NULL;
	return p;
}

void kheap_push_stack(struct kheap_mcb_t **stack, struct kheap_mcb_t *p)
{


	if(p->next)
		panic("kheap_push_stack() page is already on a stack\n");
	
	p->next = (struct kheap_mcb_t*)*stack;
	*stack = (struct kheap_mcb_t*)p;
}


int kheap_alloc_init(size_t kheap_size)
{
size_t no_pages, mcb_size,mcb_pages;
struct kheap_mcb_t *mcb;
int i;

	no_pages = KHEAP_SIZE / PAGESIZE;
	mcb_size = no_pages * sizeof(sizeof(struct kheap_mcb_t));
	mcb_pages = mcb_size / PAGESIZE;

	kheap_first_mcb = (struct kheap_mcb_t*) mm_page_alloc_cont(mcb_pages);
	if(!kheap_first_mcb)
		panic("kheap_alloc_init() unable to alloc heap data\n");

	kheap_start = (struct kheap_mcb_t*) mm_page_alloc_cont(no_pages);
	if(!kheap_start)
		panic("kheap_alloc_init() unable to alloc heap space\n");

	
	printk("kheap_alloc_init: size=%d pages=%i kheap_first_mcb=0x%p\n",mcb_size,mcb_pages,kheap_first_mcb);
	printk("kheap_alloc_init: start=0x%p no_pages=%i\n",kheap_start,no_pages);

	for(i=0 ; i < KHEAP_CHUNK_COUNT; i++)
	{
		kheap_stacks[i] = (struct kheap_mcb_t*)0;
		kheap_chunk_allocs[i] = 0;
		kheap_chunk_new_mcb_allocs[i] = 0;
	}

	for(i = no_pages ; i >= 0 ; i--)
	{
		mcb = (struct kheap_mcb_t*)&kheap_first_mcb[i];

		mcb->bitmap = 0;
		mcb->type = KHEAP_CHUNK_FREE;

		kheap_push_free_stack(mcb);
		
	}

	kheap_chunk_sizes[0] = KHEAP_CHUNK_1_SIZE;
	kheap_chunk_sizes[1] = KHEAP_CHUNK_2_SIZE;
	kheap_chunk_sizes[2] = KHEAP_CHUNK_4_SIZE;
	kheap_chunk_sizes[3] = KHEAP_CHUNK_8_SIZE;
	kheap_chunk_sizes[4] = KHEAP_CHUNK_16_SIZE;
	spinlock_init(&kheap_lock);


}

void *kmalloc(size_t size)
{
struct kheap_mcb_t *mcb;
size_t chunk_size, addr, offset;
int stack_index,i,j,bit_count,mcb_full;

	// wtf
	if(size > 0 && size <= KHEAP_CHUNK_1_SIZE)
	{
		chunk_size = KHEAP_CHUNK_1_SIZE;
		stack_index = KHEAP_CHUNK_1_INDEX;
	}
	else
	if(size > KHEAP_CHUNK_1_SIZE && size <= KHEAP_CHUNK_2_SIZE)
	{
		chunk_size = KHEAP_CHUNK_2_SIZE;
		stack_index = KHEAP_CHUNK_2_INDEX;
	}
	else
	if(size > KHEAP_CHUNK_2_SIZE && size <= KHEAP_CHUNK_4_SIZE)
	{
		chunk_size = KHEAP_CHUNK_4_SIZE;
		stack_index = KHEAP_CHUNK_4_INDEX;
	}
	else
	if(size > KHEAP_CHUNK_4_SIZE && size <= KHEAP_CHUNK_8_SIZE)
	{
		chunk_size = KHEAP_CHUNK_8_SIZE;
		stack_index = KHEAP_CHUNK_8_INDEX;
	}
	else
	if(size > KHEAP_CHUNK_8_SIZE && size <= KHEAP_CHUNK_16_SIZE)
	{
		chunk_size = KHEAP_CHUNK_16_SIZE;
		stack_index = KHEAP_CHUNK_16_INDEX;
	}
	else
		panic("kmalloc(%d) can't handle this chunk size (TODO)");

	//printk("kmalloc(%d) chunk_size=%d stack_index=%i\n",size,chunk_size,stack_index);

	// let see if there is any available chunks on the stack
	spinlock_lock(&kheap_lock);
	mcb = (struct kheap_mcb_t*) kheap_pop_stack(&kheap_stacks[stack_index]);
	if(mcb)
	{
		//printk("kmalloc() found a usable MCB %p bits %i...\n",mcb,(PAGESIZE/chunk_size));
		for(i = 0 ; i < (PAGESIZE / chunk_size) ; i ++)
		{
			if(!BitNtst(mcb->bitmap,i))
			{
				//printk("kmalloc() bit #%i is avail in %p\n",i,mcb);
				
				BitNset(mcb->bitmap,i);

				bit_count = (PAGESIZE / chunk_size);
				mcb_full = 0;
				for( j = 0 ; j < bit_count ; j++)
					BitNset(mcb_full,j);

				//printk("full mcb = %p bitmap = %p\n",mcb_full,mcb->bitmap);
		
				// If chunk not full, return to avail stack for size
				if(mcb->bitmap != mcb_full)
					kheap_push_stack(&kheap_stacks[stack_index],mcb);

				offset = ((size_t)mcb - (size_t)kheap_first_mcb) / sizeof(struct kheap_mcb_t);
				addr = (size_t)kheap_start + (offset * PAGESIZE) + (i * chunk_size);
	
				kheap_chunk_allocs[stack_index]++;
				spinlock_unlock(&kheap_lock);
				//printk("addr=%p\n",addr);
				return (void*)addr;
			}
		}

		panic("kmalloc() BUG mcb no free bit!!\n");
	}

	mcb = (struct kheap_mcb_t*)kheap_pop_free_stack();
	if(!mcb)
	{
		printk("kmalloc() out of memory!! returning NULL\n");
		spinlock_unlock(&kheap_lock);
		return NULL;
	}

	if(mcb->bitmap != 0)
		panic("kmalloc() no empty chunk on free stack");

	mcb->type = stack_index + 1;
	BitNset(mcb->bitmap,0);
	
	kheap_push_stack(&kheap_stacks[stack_index],mcb);

	offset = ((size_t)mcb - (size_t)kheap_first_mcb) / sizeof(struct kheap_mcb_t);
	addr = (size_t)kheap_start + (offset * PAGESIZE);
	
	kheap_chunk_new_mcb_allocs[stack_index]++;
	spinlock_unlock(&kheap_lock);
	return (void*)addr;
}

void kfree(void *addr)
{
struct kheap_mcb_t *mcb;
int offset;
int bit_index;

	if(!addr)
		panic("kfree() addr is NULL !!\n");

	spinlock_lock(&kheap_lock);
	offset = ((size_t)addr - (size_t)kheap_start) / PAGESIZE;
	mcb = (struct kheap_mcb_t*) &kheap_first_mcb[offset];

	bit_index = ((size_t)addr & (PAGESIZE-1)) / kheap_chunk_sizes[mcb->type-1];
	//printk("offset = %i bit_index=%i\n",offset,bit_index);

	if(!BitNtst(mcb->bitmap,bit_index))
		panic("kfree() free %p bit #%i not set!",addr,bit_index);

	// clear the bit
	BitNclr(mcb->bitmap,bit_index);
	
	// If bitmap is clear and not on an avail stack
	// push this back onto the free stack
	if(!mcb->next && mcb->bitmap == 0)
	{
		mcb->type = 0;
		kheap_push_free_stack(mcb);
	}
	else
	{
		// If bitmap was previously full
		// push to back on to the avail stack for this chunk_size
		if(!mcb->next)
			kheap_push_stack(&kheap_stacks[mcb->type-1],mcb);
		
	}

	spinlock_unlock(&kheap_lock);
	return;
}